#include "CC1101.h"

CC1101::CC1101(Module* module) : PhysicalLayer(CC1101_CRYSTAL_FREQ, CC1101_DIV_EXPONENT) {
  _mod = module;
}

int16_t CC1101::begin(float freq, float br, float rxBw, float freqDev) {
  // set module properties
  _mod->SPIreadCommand = CC1101_CMD_READ;
  _mod->SPIwriteCommand = CC1101_CMD_WRITE;
  _mod->init(USE_SPI, INT_0);
  
  // try to find the CC1101 chip
  uint8_t i = 0;
  bool flagFound = false;
  while((i < 10) && !flagFound) {
    uint8_t version = SPIreadRegister(CC1101_REG_VERSION);
    if(version == 0x14) {
      flagFound = true;
    } else {
      #ifdef RADIOLIB_DEBUG
        Serial.print(F("CC1101 not found! ("));
        Serial.print(i + 1);
        Serial.print(F(" of 10 tries) CC1101_REG_VERSION == "));
        
        char buffHex[7];
        sprintf(buffHex, "0x%04X", version);
        Serial.print(buffHex);
        Serial.print(F(", expected 0x0014"));
        Serial.println();
      #endif
      delay(1000);
      i++;
    }
  }
  
  if(!flagFound) {
    DEBUG_PRINTLN_STR("No CC1101 found!");
    SPI.end();
    return(ERR_CHIP_NOT_FOUND);
  } else {
    DEBUG_PRINTLN_STR("Found CC1101! (match by CC1101_REG_VERSION == 0x14)");
  }
  
  // configure settings not accessible by API
  int16_t state = config();
  if(state != ERR_NONE) {
    return(state);
  }
  
  // configure publicly accessible settings
  state = setFrequency(freq);
  if(state != ERR_NONE) {
    return(state);
  }
  
  state = setBitRate(br);
  if(state != ERR_NONE) {
    return(state);
  }
  
  state = setRxBandwidth(rxBw);
  if(state != ERR_NONE) {
    return(state);
  }
  
  state = setFrequencyDeviation(freqDev);
  if(state != ERR_NONE) {
    return(state);
  }
  
  // flush FIFOs
  SPIsendCommand(CC1101_CMD_FLUSH_RX);
  SPIsendCommand(CC1101_CMD_FLUSH_TX);
  
  return(state);
}

int16_t CC1101::transmit(String& str, uint8_t addr) {
  return(CC1101::transmit(str.c_str()));
}

int16_t CC1101::transmit(const char* str, uint8_t addr) {
  return(CC1101::transmit((uint8_t*)str, strlen(str)));
}

int16_t CC1101::transmit(uint8_t* data, size_t len, uint8_t addr) {
  // check packet length
  if(len > 63) {
    return(ERR_PACKET_TOO_LONG);
  }
  
  // set mode to standby
  standby();
  
  // flush Tx FIFO
  SPIsendCommand(CC1101_CMD_FLUSH_TX);
  
  // set GDO0 mapping
  int state = SPIsetRegValue(CC1101_REG_IOCFG0, CC1101_GDOX_SYNC_WORD_SENT_OR_RECEIVED);
  if(state != ERR_NONE) {
    return(state);
  }
  
  // write packet length
  SPIwriteRegister(CC1101_REG_FIFO, len);
  
  // write packet to FIFO
  SPIwriteRegisterBurst(CC1101_REG_FIFO, data, len);
  
  // set mode to transmit
  SPIsendCommand(CC1101_CMD_TX);
  
  // wait for transmission start
  while(!digitalRead(_mod->getInt0()));
  
  // wait for transmission end
  while(digitalRead(_mod->getInt0()));
  
  // set mode to standby
  standby();
  
  // flush Tx FIFO
  SPIsendCommand(CC1101_CMD_FLUSH_TX);
  
  return(state);
}

int16_t CC1101::receive(String& str, size_t len) {
  // create temporary array to store received data
  char* data = new char[len + 1];
  int16_t state = CC1101::receive((uint8_t*)data, len);
  
  // if packet was received successfully, copy data into String
  if(state == ERR_NONE) {
    str = String(data);
  }
  
  delete[] data;
  return(state);
}

int16_t CC1101::receive(uint8_t* data, size_t len) {
  // set mode to standby
  standby();
  
  // flush Rx FIFO
  SPIsendCommand(CC1101_CMD_FLUSH_RX);
  
  // set GDO0 mapping
  int state = SPIsetRegValue(CC1101_REG_IOCFG0, CC1101_GDOX_SYNC_WORD_SENT_OR_RECEIVED);
  if(state != ERR_NONE) {
    return(state);
  }
  
  // set mode to receive
  SPIsendCommand(CC1101_CMD_RX);
  
  // wait for sync word
  while(!digitalRead(_mod->getInt0()));
  
  // wait for packet end
  while(digitalRead(_mod->getInt0()));
  
  // get packet length
  size_t length = SPIreadRegister(CC1101_REG_RXBYTES) - 2;

  // read packet data
  if(len == 0) {
    // argument len equal to zero indicates String call, which means dynamically allocated data array
    // dispose of the original and create a new one
    delete[] data;
    data = new uint8_t[length + 1];
  }
  SPIreadRegisterBurst(CC1101_REG_FIFO, length, data);
  
  // read RSSI byte
  _rawRSSI = SPIgetRegValue(CC1101_REG_FIFO);
  
  // read LQI and CRC byte
  uint8_t val = SPIgetRegValue(CC1101_REG_FIFO);
  _rawLQI = val & 0x7F;
  
  // add terminating null
  if(len == 0) {
    data[length] = 0;
  }
  
  // flush Rx FIFO
  SPIsendCommand(CC1101_CMD_FLUSH_RX);
  
  // set mode to standby
  standby();
  
  // check CRC
  if((val & 0b10000000) == 0b00000000) {
    return(ERR_CRC_MISMATCH);
  }
  
  return(state);
}

int16_t CC1101::standby() {
  SPIsendCommand(CC1101_CMD_IDLE);
  return(ERR_NONE);
}

int16_t CC1101::transmitDirect(uint32_t FRF) {
  // user requested to start transmitting immediately (required for RTTY)
  if(FRF != 0) {
    SPIwriteRegister(CC1101_REG_FREQ2, (FRF & 0xFF0000) >> 16);
    SPIwriteRegister(CC1101_REG_FREQ1, (FRF & 0x00FF00) >> 8);
    SPIwriteRegister(CC1101_REG_FREQ0, FRF & 0x0000FF);
  
    SPIsendCommand(CC1101_CMD_TX);
  }
  
  // activate direct mode
  int16_t state = directMode();
  if(state != ERR_NONE) {
    return(state);
  }
  
  // start transmitting
  SPIsendCommand(CC1101_CMD_TX);
  return(state);
}

int16_t CC1101::receiveDirect() {
  // activate direct mode
  int16_t state = directMode();
  if(state != ERR_NONE) {
    return(state);
  }
  
  // start receiving
  SPIsendCommand(CC1101_CMD_RX);
  return(ERR_NONE);
}

void CC1101::setGdo0Action(void (*func)(void), uint8_t dir) {
  attachInterrupt(digitalPinToInterrupt(_mod->getInt0()), func, dir);
}

void CC1101::setGdo1Action(void (*func)(void), uint8_t dir) {
  attachInterrupt(digitalPinToInterrupt(_mod->getInt0()), func, dir);
}

int16_t CC1101::startTransmit(String& str, uint8_t addr) {
  return(CC1101::startTransmit(str.c_str(), addr));
}

int16_t CC1101::startTransmit(const char* str, uint8_t addr) {
  return(CC1101::startTransmit((uint8_t*)str, strlen(str), addr));
}

int16_t CC1101::startTransmit(uint8_t* data, size_t len, uint8_t addr) {
  // check packet length
  if(len > 63) {
    return(ERR_PACKET_TOO_LONG);
  }
  
  // set mode to standby
  standby();
  
  // flush Tx FIFO
  SPIsendCommand(CC1101_CMD_FLUSH_TX);
  
  // set GDO0 mapping
  int state = SPIsetRegValue(CC1101_REG_IOCFG0, CC1101_GDOX_SYNC_WORD_SENT_OR_RECEIVED);
  if(state != ERR_NONE) {
    return(state);
  }
  
  // write packet length
  SPIwriteRegister(CC1101_REG_FIFO, len);
  
  // write packet to FIFO
  SPIwriteRegisterBurst(CC1101_REG_FIFO, data, len);
  
  // set mode to transmit
  SPIsendCommand(CC1101_CMD_TX);
  
  return(state);
}

int16_t CC1101::startReceive() {
  // set mode to standby
  standby();
  
  // flush Rx FIFO
  SPIsendCommand(CC1101_CMD_FLUSH_RX);
  
  // set GDO0 mapping
  int state = SPIsetRegValue(CC1101_REG_IOCFG0, CC1101_GDOX_SYNC_WORD_SENT_OR_RECEIVED);
  if(state != ERR_NONE) {
    return(state);
  }
  
  // set mode to receive
  SPIsendCommand(CC1101_CMD_RX);
  
  return(state);
}

int16_t CC1101::readData(String& str, size_t len) {
  // create temporary array to store received data
  char* data = new char[len];
  int16_t state = CC1101::readData((uint8_t*)data, len);
  
  // if packet was received successfully, copy data into String
  if(state == ERR_NONE) {
    str = String(data);
  }
  
  delete[] data;
  return(state);
}

int16_t CC1101::readData(uint8_t* data, size_t len) {
  // get packet length
  size_t length = SPIreadRegister(CC1101_REG_RXBYTES) - 2;

  // read packet data
  if(len == 0) {
    // argument len equal to zero indicates String call, which means dynamically allocated data array
    // dispose of the original and create a new one
    delete[] data;
    data = new uint8_t[length + 1];
  }
  SPIreadRegisterBurst(CC1101_REG_FIFO, length, data);
  
  // read RSSI byte
  _rawRSSI = SPIgetRegValue(CC1101_REG_FIFO);
  
  // read LQI and CRC byte
  uint8_t val = SPIgetRegValue(CC1101_REG_FIFO);
  _rawLQI = val & 0x7F;
  
  // add terminating null
  if(len == 0) {
    data[length] = 0;
  }
  
  // flush Rx FIFO
  SPIsendCommand(CC1101_CMD_FLUSH_RX);
  
  // set mode to receive
  SPIsendCommand(CC1101_CMD_RX);
  
  // check CRC
  if((val & 0b10000000) == 0b00000000) {
    return(ERR_CRC_MISMATCH);
  }
  
  return(ERR_NONE);
}


int16_t CC1101::setFrequency(float freq) {
  // check allowed frequency range
  if(!(((freq > 300.0) && (freq < 348.0)) ||
       ((freq > 387.0) && (freq < 464.0)) ||
       ((freq > 779.0) && (freq < 928.0)))) {
    return(ERR_INVALID_FREQUENCY);
  }
  
  // set mode to standby
  SPIsendCommand(CC1101_CMD_IDLE);
  
  //set carrier frequency
  uint32_t base = 1;
  uint32_t FRF = (freq * (base << 16)) / 26.0;
  int16_t state = SPIsetRegValue(CC1101_REG_FREQ2, (FRF & 0xFF0000) >> 16, 7, 0);
  state |= SPIsetRegValue(CC1101_REG_FREQ1, (FRF & 0x00FF00) >> 8, 7, 0);
  state |= SPIsetRegValue(CC1101_REG_FREQ0, FRF & 0x0000FF, 7, 0);

  return(state);
}

int16_t CC1101::setBitRate(float br) {
  // check allowed bit rate range
  if(!((br >= 0.025) && (br <= 600.0))) {
    return(ERR_INVALID_BIT_RATE);
  }
  
  // set mode to standby
  SPIsendCommand(CC1101_CMD_IDLE);
  
  // calculate exponent and mantisa values
  uint8_t e = 0;
  uint8_t m = 0;
  getExpMant(br * 1000.0, 256, 28, 14, e, m);
  
  // set bit rate value
  int16_t state = SPIsetRegValue(CC1101_REG_MDMCFG4, e, 3, 0);
  state |= SPIsetRegValue(CC1101_REG_MDMCFG3, m);
  return(state);
}

int16_t CC1101::setRxBandwidth(float rxBw) {
  // check allowed bandwidth range
  if(!((rxBw >= 58) && (rxBw <= 812))) {
    return(ERR_INVALID_RX_BANDWIDTH);
  }
  
  // set mode to standby
  SPIsendCommand(CC1101_CMD_IDLE);
  
  // calculate exponent and mantisa values
  for(int8_t e = 3; e >= 0; e--) {
    for(int8_t m = 3; m >= 0; m --) {
      float point = (CC1101_CRYSTAL_FREQ * 1000000.0)/(8 * (m + 4) * ((uint32_t)1 << e));
      if(abs((rxBw * 1000.0) - point) <= 0.001) {
        // set Rx channel filter bandwidth
        return(SPIsetRegValue(CC1101_REG_MDMCFG4, (e << 6) | (m << 4), 7, 4));
      }
    }
  }
  
  return(ERR_UNKNOWN);
}

int16_t CC1101::setFrequencyDeviation(float freqDev) {
  // set frequency deviation to lowest available setting (required for RTTY)
  if(freqDev == 0.0) {
    int16_t state = SPIsetRegValue(CC1101_REG_DEVIATN, 0, 6, 4);
    state |= SPIsetRegValue(CC1101_REG_DEVIATN, 0, 2, 0);
    return(state);
  }

  // check allowed frequency deviation range
  if(!((freqDev >= 1.587) && (freqDev <= 380.8))) {
    return(ERR_INVALID_FREQUENCY_DEVIATION);
  }
  
  // set mode to standby
  SPIsendCommand(CC1101_CMD_IDLE);
  
  // calculate exponent and mantisa values
  uint8_t e = 0;
  uint8_t m = 0;
  getExpMant(freqDev * 1000.0, 8, 17, 7, e, m);
  
  // set frequency deviation value
  int16_t state = SPIsetRegValue(CC1101_REG_DEVIATN, (e << 4), 6, 4);
  state |= SPIsetRegValue(CC1101_REG_DEVIATN, m, 2, 0);
  return(state);
}

int16_t CC1101::setSyncWord(uint8_t syncH, uint8_t syncL) {
  // set sync word
  int16_t state = SPIsetRegValue(CC1101_REG_SYNC1, syncH);
  state |= SPIsetRegValue(CC1101_REG_SYNC0, syncL);
  return(state);
}

int16_t CC1101::setNodeAddress(uint8_t nodeAddr, uint8_t numBroadcastAddrs) {
  if(!(numBroadcastAddrs > 0) && (numBroadcastAddrs <= 2)) {
    return(ERR_INVALID_NUM_BROAD_ADDRS);
  }

  // enable address filtering
  int16_t state = SPIsetRegValue(CC1101_REG_PKTCTRL1, numBroadcastAddrs + 0x01, 1, 0);
  if(state != ERR_NONE) {
    return(state);
  }
  
  // set node address
  return(SPIsetRegValue(CC1101_REG_ADDR, nodeAddr));
}

int16_t CC1101::disableAddressFiltering() {
  // disable address filtering
  int16_t state = _mod->SPIsetRegValue(CC1101_REG_PKTCTRL1, CC1101_ADR_CHK_NONE, 1, 0);
  if(state != ERR_NONE) {
    return(state);
  }
  
  // set node address to default (0x00)
  return(SPIsetRegValue(CC1101_REG_ADDR, 0x00));
}

float CC1101::getRSSI() {
  float rssi;
  if(_rawRSSI >= 128) {
    rssi = (((float)_rawRSSI - 256.0)/2.0) - 74.0;
  } else {
    rssi = (((float)_rawRSSI)/2.0) - 74.0;
  }
  return(rssi);
}

uint8_t CC1101::getLQI() {
  return(_rawLQI);
}

int16_t CC1101::config() {
  // enable automatic frequency synthesizer calibration 
  int16_t state = SPIsetRegValue(CC1101_REG_MCSM0, CC1101_FS_AUTOCAL_IDLE_TO_RXTX, 5, 4);
  if(state != ERR_NONE) {
    return(state);
  }
  
  // set packet mode
  state = SPIsetRegValue(CC1101_REG_PKTCTRL1, CC1101_CRC_AUTOFLUSH_OFF | CC1101_APPEND_STATUS_ON | CC1101_ADR_CHK_NONE, 3, 0); 
  state |= SPIsetRegValue(CC1101_REG_PKTCTRL0, CC1101_WHITE_DATA_OFF | CC1101_PKT_FORMAT_NORMAL, 6, 4);
  state |= SPIsetRegValue(CC1101_REG_PKTCTRL0, CC1101_CRC_ON | CC1101_LENGTH_CONFIG_VARIABLE, 2, 0);
  if(state != ERR_NONE) {
    return(state);
  }
  
  // TODO: configurable power output
  SPIwriteRegister(CC1101_REG_PATABLE, 0x60);
  
  return(state);
}

int16_t CC1101::directMode() {
  // set mode to standby
  SPIsendCommand(CC1101_CMD_IDLE);
  
  // set GDO0 and GDO2 mapping
  int16_t state = SPIsetRegValue(CC1101_REG_IOCFG0, CC1101_GDOX_SERIAL_CLOCK , 5, 0);
  state |= SPIsetRegValue(CC1101_REG_IOCFG2, CC1101_GDOX_SERIAL_DATA_SYNC , 5, 0);
  
  // set continuous mode
  state |= SPIsetRegValue(CC1101_REG_PKTCTRL0, CC1101_PKT_FORMAT_SYNCHRONOUS, 5, 4);
  return(state);
}

void CC1101::getExpMant(float target, uint16_t mantOffset, uint8_t divExp, uint8_t expMax, uint8_t& exp, uint8_t& mant) {
  // get table origin point (exp = 0, mant = 0)
  float origin = (mantOffset * CC1101_CRYSTAL_FREQ * 1000000.0)/((uint32_t)1 << divExp);
  
  // iterate over possible exponent values
  for(int8_t e = expMax; e >= 0; e--) {
    // get table column start value (exp = e, mant = 0);
	  float intervalStart = ((uint32_t)1 << e) * origin;
    
    // check if target value is in this column
	  if(target >= intervalStart) {
      // save exponent value
      exp = e;
      
      // calculate size of step between table rows
	    float stepSize = intervalStart/(float)mantOffset;
      
      // get target point position (exp = e, mant = m)
	    mant = ((target - intervalStart) / stepSize);
      
      // we only need the first match, terminate
	    return;
	  }
	}
}



int16_t CC1101::SPIgetRegValue(uint8_t reg, uint8_t msb, uint8_t lsb) {
  // status registers require special command
  if(reg > CC1101_REG_TEST0) {
    reg |= CC1101_CMD_ACCESS_STATUS_REG;
  }
  
  return(_mod->SPIgetRegValue(reg, msb, lsb));
}

int16_t CC1101::SPIsetRegValue(uint8_t reg, uint8_t value, uint8_t msb, uint8_t lsb, uint8_t checkInterval) {
  // status registers require special command
  if(reg > CC1101_REG_TEST0) {
    reg |= CC1101_CMD_ACCESS_STATUS_REG;
  }
  
  return(_mod->SPIsetRegValue(reg, value, msb, lsb, checkInterval));
}

void CC1101::SPIreadRegisterBurst(uint8_t reg, uint8_t numBytes, uint8_t* inBytes) {
  _mod->SPIreadRegisterBurst(reg | CC1101_CMD_BURST, numBytes, inBytes);
}

uint8_t CC1101::SPIreadRegister(uint8_t reg) {
  // status registers require special command
  if(reg > CC1101_REG_TEST0) {
    reg |= CC1101_CMD_ACCESS_STATUS_REG;
  }

  return(_mod->SPIreadRegister(reg));
}

void CC1101::SPIwriteRegister(uint8_t reg, uint8_t data) {
  // status registers require special command
  if(reg > CC1101_REG_TEST0) {
    reg |= CC1101_CMD_ACCESS_STATUS_REG;
  }
  
  return(_mod->SPIwriteRegister(reg, data));
}

void CC1101::SPIwriteRegisterBurst(uint8_t reg, uint8_t* data, size_t len) {
  _mod->SPIwriteRegisterBurst(reg | CC1101_CMD_BURST, data, len);
}

void CC1101::SPIsendCommand(uint8_t cmd) {
  digitalWrite(_mod->getCs(), LOW);
  SPI.beginTransaction(SPISettings(2000000, MSBFIRST, SPI_MODE0));
  SPI.transfer(cmd);
  SPI.endTransaction();
  digitalWrite(_mod->getCs(), HIGH);
}
